! ***********************************************************************
!
!   Copyright (C) 2012  Bill Paxton
!
!   this file is part of mesa.
!
!   mesa is free software; you can redistribute it and/or modify
!   it under the terms of the gnu general library public license as published
!   by the free software foundation; either version 2 of the license, or
!   (at your option) any later version.
!
!   mesa is distributed in the hope that it will be useful, 
!   but without any warranty; without even the implied warranty of
!   merchantability or fitness for a particular purpose.  see the
!   gnu library general public license for more details.
!
!   you should have received a copy of the gnu library general public license
!   along with this software; if not, write to the free software
!   foundation, inc., 59 temple place, suite 330, boston, ma 02111-1307 usa
!
! ***********************************************************************
 
      module run_star_extras 

      use star_lib
      use star_def
      use const_def
      use chem_def
      use num_lib
      use binary_def
      
      implicit none

      real(dp) :: xfer_old = 1.0

      ! these routines are called by the standard run_star check_model
      contains
      
      subroutine extras_controls(id, ierr)
         integer, intent(in) :: id
         integer, intent(out) :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         
         ! this is the place to set any procedure pointers you want to change
         ! e.g., other_wind, other_mixing, other_energy  (see star_data.inc)
         
         ! Uncomment these lines if you wish to use the functions in this file,
         ! otherwise we use a null_ version which does nothing.
         s% extras_startup => extras_startup
         s% extras_check_model => extras_check_model
         s% extras_finish_step => extras_finish_step
         s% extras_after_evolve => extras_after_evolve
         s% how_many_extra_history_columns => how_many_extra_history_columns
         s% data_for_extra_history_columns => data_for_extra_history_columns
         s% how_many_extra_profile_columns => how_many_extra_profile_columns
         s% data_for_extra_profile_columns => data_for_extra_profile_columns  

         ! Once you have set the function pointers you want,
         ! then uncomment this (or set it in your star_job inlist)
         ! to disable the printed warning message,
          s% job% warn_run_star_extras =.false.       
            
          s% other_wind => rlo_other_wind

      end subroutine extras_controls
      
      ! None of the following functions are called unless you set their
      ! function point in extras_control.
      
      
      integer function extras_startup(id, restart, ierr)
         integer, intent(in) :: id
         logical, intent(in) :: restart
         integer, intent(out) :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         extras_startup = 0
         if (.not. restart) then
            call alloc_extra_info(s)
         else ! it is a restart
            call unpack_extra_info(s)
         end if
      end function extras_startup
      

      ! returns either keep_going, retry, backup, or terminate.
      integer function extras_check_model(id, id_extra)
         integer, intent(in) :: id, id_extra
         integer :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         extras_check_model = keep_going         
         if (.false. .and. s% star_mass_h1 < 0.35d0) then
            ! stop when star hydrogen mass drops to specified level
            extras_check_model = terminate
            write(*, *) 'have reached desired hydrogen mass'
            return
         end if


         ! if you want to check multiple conditions, it can be useful
         ! to set a different termination code depending on which
         ! condition was triggered.  MESA provides 9 customizeable
         ! termination codes, named t_xtra1 .. t_xtra9.  You can
         ! customize the messages that will be printed upon exit by
         ! setting the corresponding termination_code_str value.
         ! termination_code_str(t_xtra1) = 'my termination condition'

         ! by default, indicate where (in the code) MESA terminated
         if (extras_check_model == terminate) s% termination_code = t_extras_check_model
      end function extras_check_model


      integer function how_many_extra_history_columns(id, id_extra)
         integer, intent(in) :: id, id_extra
         integer :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         how_many_extra_history_columns = 0
      end function how_many_extra_history_columns
      
      
      subroutine data_for_extra_history_columns(id, id_extra, n, names, vals, ierr)
         integer, intent(in) :: id, id_extra, n
         character (len=maxlen_history_column_name) :: names(n)
         real(dp) :: vals(n)
         integer, intent(out) :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         
         !note: do NOT add the extras names to history_columns.list
         ! the history_columns.list is only for the built-in log column options.
         ! it must not include the new column names you are adding here.
         

      end subroutine data_for_extra_history_columns

      
      integer function how_many_extra_profile_columns(id, id_extra)
         use star_def, only: star_info
         integer, intent(in) :: id, id_extra
         integer :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         how_many_extra_profile_columns = 0
      end function how_many_extra_profile_columns
      
      
      subroutine data_for_extra_profile_columns(id, id_extra, n, nz, names, vals, ierr)
         use star_def, only: star_info, maxlen_profile_column_name
         use const_def, only: dp
         integer, intent(in) :: id, id_extra, n, nz
         character (len=maxlen_profile_column_name) :: names(n)
         real(dp) :: vals(nz,n)
         integer, intent(out) :: ierr
         type (star_info), pointer :: s
         integer :: k
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         
         !note: do NOT add the extra names to profile_columns.list
         ! the profile_columns.list is only for the built-in profile column options.
         ! it must not include the new column names you are adding here.

         ! here is an example for adding a profile column
         !if (n /= 1) stop 'data_for_extra_profile_columns'
         !names(1) = 'beta'
         !do k = 1, nz
         !   vals(k,1) = s% Pgas(k)/s% P(k)
         !end do
         
      end subroutine data_for_extra_profile_columns
      

      ! returns either keep_going or terminate.
      ! note: cannot request retry or backup; extras_check_model can do that.
      integer function extras_finish_step(id, id_extra)
         integer, intent(in) :: id, id_extra
         integer :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         extras_finish_step = keep_going
         call store_extra_info(s)

         ! to save a profile, 
            ! s% need_to_save_profiles_now = .true.
         ! to update the star log,
            ! s% need_to_update_history_now = .true.

         ! see extras_check_model for information about custom termination codes
         ! by default, indicate where (in the code) MESA terminated
         if (extras_finish_step == terminate) s% termination_code = t_extras_finish_step
      end function extras_finish_step
      
      
      subroutine extras_after_evolve(id, id_extra, ierr)
         integer, intent(in) :: id, id_extra
         integer, intent(out) :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
      end subroutine extras_after_evolve
      
      
      ! routines for saving and restoring extra data so can do restarts
         
         ! put these defs at the top and delete from the following routines
         !integer, parameter :: extra_info_alloc = 1
         !integer, parameter :: extra_info_get = 2
         !integer, parameter :: extra_info_put = 3
      
      
      subroutine alloc_extra_info(s)
         integer, parameter :: extra_info_alloc = 1
         type (star_info), pointer :: s
         call move_extra_info(s,extra_info_alloc)
      end subroutine alloc_extra_info
      
      
      subroutine unpack_extra_info(s)
         integer, parameter :: extra_info_get = 2
         type (star_info), pointer :: s
         call move_extra_info(s,extra_info_get)
      end subroutine unpack_extra_info
      
      
      subroutine store_extra_info(s)
         integer, parameter :: extra_info_put = 3
         type (star_info), pointer :: s
         call move_extra_info(s,extra_info_put)
      end subroutine store_extra_info
      
      
      subroutine move_extra_info(s,op)
         integer, parameter :: extra_info_alloc = 1
         integer, parameter :: extra_info_get = 2
         integer, parameter :: extra_info_put = 3
         type (star_info), pointer :: s
         integer, intent(in) :: op
         
         integer :: i, j, num_ints, num_dbls, ierr
         
         i = 0
         ! call move_int or move_flg    
         num_ints = i
         
         i = 0
         ! call move_dbl       
         
         num_dbls = i
         
         if (op /= extra_info_alloc) return
         if (num_ints == 0 .and. num_dbls == 0) return
         
         ierr = 0
         call star_alloc_extras(s% id, num_ints, num_dbls, ierr)
         if (ierr /= 0) then
            write(*,*) 'failed in star_alloc_extras'
            write(*,*) 'alloc_extras num_ints', num_ints
            write(*,*) 'alloc_extras num_dbls', num_dbls
            stop 1
         end if
         
         contains
         
         subroutine move_dbl(dbl)
            real(dp) :: dbl
            i = i+1
            select case (op)
            case (extra_info_get)
               dbl = s% extra_work(i)
            case (extra_info_put)
               s% extra_work(i) = dbl
            end select
         end subroutine move_dbl
         
         subroutine move_int(int)
            integer :: int
            i = i+1
            select case (op)
            case (extra_info_get)
               int = s% extra_iwork(i)
            case (extra_info_put)
               s% extra_iwork(i) = int
            end select
         end subroutine move_int
         
         subroutine move_flg(flg)
            logical :: flg
            i = i+1
            select case (op)
            case (extra_info_get)
               flg = (s% extra_iwork(i) /= 0)
            case (extra_info_put)
               if (flg) then
                  s% extra_iwork(i) = 1
               else
                  s% extra_iwork(i) = 0
               end if
            end select
         end subroutine move_flg
      
      end subroutine move_extra_info

      subroutine rlo_other_wind(id, Lsurf, Msurf, Rsurf, Tsurf, w, ierr)
         use star_def
         integer, intent(in) :: id
         real(dp), intent(in) :: Lsurf, Msurf, Rsurf, Tsurf ! surface values (cgs)
         ! NOTE: surface is outermost cell. not necessarily at photosphere.
         ! NOTE: don't assume that vars are set at this point.
         ! so if you want values other than those given as args,
         ! you should use values from s% xh(:,:) and s% xa(:,:) only.
         ! rather than things like s% Teff or s% lnT(:) which have not been set yet.
         real(dp), intent(out) :: w ! wind in units of Msun/year (value is >= 0)
         integer, intent(out) :: ierr
         integer :: i
         real(dp) :: roche_lobe_radius ! Rsun
         real(dp) :: ratio, xfer_ratio
         real(dp) :: roche_lobe_xfer_full_on, roche_lobe_xfer_full_off,&
            rlo_wind_base_mdot, rlo_wind_scale_height, rlo_wind_on
         real(dp) :: vx, rlobe2, donor_mass, prev_mdot_frac
         real(dp) :: radius, radius_xfer_full_on, radius_xfer_full_off

         include 'formats'

         type (star_info), pointer :: s
         type (binary_info), pointer :: b

         call binary_ptr(1, b,ierr)
         call star_ptr(1, s, ierr)

         donor_mass = s% star_mass
         call star_ptr(2, s, ierr)

         vx = (s% star_mass/donor_mass)**one_third

         ! Eggleton 1983 Estimate
         rlobe2 = b% separation*0.49d0*vx*vx/(0.6d0*vx*vx + log(1d0 + vx))
         radius = (10**(s% log_surface_radius)*Rsun )
         ratio = radius/rlobe2

         ! No mass transfer below this radius
         radius_xfer_full_on = s% x_ctrl(1)*Rsun ! 0.08

         ! "Full" mass transfer above this radius
         radius_xfer_full_off = s% x_ctrl(2)*Rsun ! 0.50

         ! Assume 100% transfer efficiency for now, will adjust later
         xfer_ratio = 1d0

         ! If we're in between the two critical ratios, interpolate xfer_ratio
         ! between 1 (full_on) and 0 (full_off)
         if (radius > radius_xfer_full_on .and. radius < radius_xfer_full_off) then
            ! linear scaling, BILL WOLF added this
            xfer_ratio = 1-(radius - radius_xfer_full_on) / &
               (radius_xfer_full_off - radius_xfer_full_on)
            xfer_ratio = 0.5d0*(1 - cos(pi*xfer_ratio))! This is a nice sinusoidal turn-on and turn-off

            !xfer_ratio = (1-((ratio-roche_lobe_xfer_full_on) / &
               !(roche_lobe_xfer_full_off-roche_lobe_xfer_full_on))**2)**(1.0/2.0) ! This one starts changing slowly at the lower radius, but faster at the upper. I thought it would help at some point, but I'm not sure it makes a difference
         endif

         ! No transfer if radius is too big
         if (radius > radius_xfer_full_off) then
            xfer_ratio = 0d0
         endif
         write(*,*) radius/radius_xfer_full_off

         ! Implicit: full transfer if radius is below roche_lobe_xfer_full_on

         write(*,*) 'xfer_ratio', xfer_ratio, 'period', b% period/60d0
         write(*,*) 'ratio', ratio
         ! To prevent wild swings, heavily weight xfer_fraction by previous
         ! value (which is thus heavily weighted by subsequently previous
         ! xfer_ratios). You may want to change this to something smaller.
         prev_mdot_frac = 0.95
!         xfer_ratio = prev_mdot_frac * xfer_old + &
!            (1.0 - prev_mdot_frac) * xfer_ratio
!         write(*,*) 'xfer_old', xfer_old, 'xfer', xfer_ratio
!         xfer_old = xfer_ratio


         ! mass_transfer_beta is the fractional amount of mass transfer that is
         ! LOST from the vicinity of the accretor (I think), so
         ! xfer_ration = 1 => mass_transfer_beta = 0 => fully conservative
         ! mass transfer
         b% mass_transfer_beta = 1d0-xfer_ratio

!         if (-b% mtransfer_rate*(secyer/msun) > 3d-7) then
!            b% mass_transfer_beta = 1+3d-7/(b% mtransfer_rate*(secyer/msun))
!         endif

         w = 0
         ierr = 0

         ! Now for the case of ratio > 1
         ! Use rlo inlist parameters, leaving rlo_scaling_factor = 0
         ! need to set rlo_wind_scale_height
         ! need to set rlo_wind_base_mdot

         rlo_wind_base_mdot = s% x_ctrl(5)
         rlo_wind_scale_height = s% x_ctrl(4)

         ! The ratio above which net mass LOSS occurs (in addition to zero
         ! transfer)
         rlo_wind_on = s% x_ctrl(3) !0.9

         if (ratio > rlo_wind_on) then

            ! Normal exponential growth of wind
            w = rlo_wind_base_mdot * &
               exp((10 ** (s% log_surface_radius) * Rsun - rlobe2) / &
               (Rsun * rlo_wind_scale_height))

            ! OVERWRITE previous mass_transfer_beta and make it be > 1, meaning
            ! more mass is being lost than was originially thought to be
            ! transferred. This looks wonky but I think that in this convention,
            ! b% mtransfer_rate is actually negative
            ! This is because just setting w doesn't work in binary for whatever reason
            ! so you have to set mass_transfer_beta negative, as a value of 0 says keep all the mass,
            ! 1 says keep none of the mass, so >1 means lose mass
            ! I don't it was meant to be used this way, but it works
            b% mass_transfer_beta = 1d0
            b% mass_transfer_beta = 1-(msun/secyer)*w/(-1.5051355143841392D+20)!b% mtransfer_rate


            write(*,1) 'log rlo mdot Msun/yr, R/R_L, beta', log10(w), ratio, b% mass_transfer_beta
         endif

         ! Some diagnostics you could output
         if (.false. .and. (Rsurf/Rsun - roche_lobe_radius)/rlo_wind_scale_height > 100) then
            write(*,1) 'ratio', (Rsurf/Rsun - roche_lobe_radius)/rlo_wind_scale_height
            write(*,1) 'rlo_wind_scale_height', rlo_wind_scale_height
            write(*,1) 'R - roche_lobe_radius', Rsurf/Rsun - roche_lobe_radius
            write(*,1) 'R', Rsurf/Rsun
            write(*,1) 'exp(s% xh(s% i_lnR,1))/Rsun', &
               exp(s% xh(s% i_lnR,1))/Rsun
            write(*,1) 'roche_lobe_radius', roche_lobe_radius
            stop 'rlo_other_wind'
         end if



      end subroutine rlo_other_wind


      end module run_star_extras
      
